body {
  /* to match ua.css, see bug 1020143 */
  -moz-font-feature-settings: "tnum";
  font-feature-settings: "tnum";
}
div, p {
  padding: 0; margin: 0;
  line-height: 150%;
  float: left;
}
p {
  padding-right: .5em;
}
